import datetime
from dataclasses import dataclass


@dataclass
class IntervalOffset:
    interval: datetime.timedelta
    offset: datetime.datetime = datetime.datetime(
        2000, 1, 1, tzinfo=datetime.timezone.utc
    )

    def get_intervals(self):
        now = datetime.datetime.now(datetime.timezone.utc)
        print(now)
        print(self.offset)
        delta = now - self.offset
        print(delta)
        return delta // self.interval



import hashlib


class Hasher:
    """
    Simple class to handle hashing of file data and strings.
    """

    def __init__(self):
        self.hash = hashlib.sha256()

    def add_file(self, path):
        self.add_str(str(path))
        with open(path, "rb") as f:
            self.hash.update(f.read())

    def add_str(self, string):
        self.hash.update(string.encode("utf-8"))

    def hexdigest(self):
        return self.hash.hexdigest()



import enum
import subprocess
from dataclasses import dataclass
from functools import cached_property
from typing import Optional

from .buildconfig import BuildConfig


class RemotePolicy(enum.Enum):
    """
    Defines how the remote server is used.
    """

    ALL = (
        enum.auto()
    )  # The server will be queried before building and the built image will be pushed
    PULL_ONLY = enum.auto()  # The build image will not be pushed
    PUSH_ONLY = enum.auto()  # The server will not be queried for the image
    NONE = enum.auto()  # The server will not be used for anything


@dataclass
class DockerImage:
    """

    Params:
    name: The name of the image
    build_config:  Config to describe how to build the image
    with_hash: If true the build config state hash will be appended to the tag
    force_build: Force building of the image even if it already exists

    version: Optional version string to add to the tag
    hash_len: Length of the hash to append if ` with_hash` is True. Default is 16

    registry:  DockerRegistry object for builds using a remote server
    prepend_server: If true, the server url will be prepended to the image name before pushing. Set this to false
        if you have already prepended the registry to the image name (e.g. docker.io/image_name)
    remote_policy:  How the remote server will be used to push and pull images
    """

    name: str
    build_config: BuildConfig = None
    with_hash: bool = False
    force_build: bool = False

    version: Optional[str] = None
    hash_len: int = 16

    registry: Optional["DockerRegistry"] = None
    prepend_server: bool = True
    remote_policy: RemotePolicy = RemotePolicy.ALL

    class BuildFailedException(Exception):
        pass

    class UnhashableReferenceException(Exception):
        pass

    def __post_init__(self):
        if self.with_hash:
            if self.build_config is None:
                raise DockerImage.UnhashableReferenceException(
                    f"The image {self.name} is configured to have a state hash appended to its tag, but the build config is None. "
                    "Please disable appending the hash on the tag (with_hash = False) or add a build config."
                )

            if not self.build_config.is_hashable():
                raise DockerImage.UnhashableReferenceException(
                    f"The image {self.name} is configured to have a state hash appended to its tag, but the config is unhashable. "
                    "Please disable appending the hash on the tag (with_hash = False) or change the build config to make it hashable."
                )

    def has_local_image(self):
        """
        Check if the image exists locally
        """
        p = subprocess.run(
            ["docker", "image", "inspect", self.ref], stdout=subprocess.DEVNULL
        )
        return p.returncode == 0

    @cached_property
    def reference(self):
        """
        Returns the reference string (e.g. image:tag-1.0-ab01) for the image using a hash of its dependencies.
        """

        tag_parts = []

        if self.version:
            tag_parts.append(self.version)

        if self.with_hash:
            tag_parts.append(self.build_config.get_hash()[: self.hash_len])

        if not tag_parts:
            return self.name

        return self.name + ":" + "-".join(tag_parts)

    @property
    def ref(self):
        return self.reference

    @cached_property
    def registry_reference(self):
        """
        Returns the registry prepended to the reference string (e.g. docker.io/image:tag).
        """

        return self.registry.prepend_server(self.reference)

    def check_existence(self):
        if self.has_local_image():
            print("<<< Image already exists locally <<<")
            return True

        if self.registry and self.remote_policy in {
            RemotePolicy.ALL,
            RemotePolicy.PULL_ONLY,
        }:
            print("Checking for image on server...")
            if self.registry.try_pull_image(self.ref, self.prepend_server):
                print("<<< Pulled image from server <<<")
                return True

        return False

    def ensure(self):
        """
        Ensures that the image is available on the local system.
        By the time this function returns, the image will exist. It will be downloaded or built if necessary.
        """
        print(f">>> Ensuring image {self.ref} >>>")

        if not self.force_build and self.check_existence():
            return

        if self.build_config is None:
            raise DockerImage.BuildFailedException(
                f"Image {self.ref} needs to be built but it has no build config, so it can't be ensured."
            )

        print(f"Building {self.ref}")
        self.build_config.build_image(self.ref)

        if self.registry and self.remote_policy in {
            RemotePolicy.ALL,
            RemotePolicy.PUSH_ONLY,
        }:
            print("Pushing image")
            self.registry.push_image(self.ref, self.prepend_server)

        print("<<< Built image <<<")

        return



import subprocess
from dataclasses import dataclass
from typing import Optional

from .image import DockerImage


@dataclass
class DockerRegistry:
    """
    Provides functions to interact with a remote Docker server: logging in, pushing and pulling images.
    """

    server: Optional[str]  # Set to None for the default Docker registry
    username: Optional[str] = None
    password: Optional[str] = None

    def __post_init__(self):
        self.loggedin = False

    def login(self):
        if self.loggedin:
            return

        if not self.username:
            return

        self.loggedin = True

        args = ["docker", "login"]
        if self.server:
            args += [self.server]
        args += ["-u", self.username, "-p", self.password]
        subprocess.run(args, check=True)

    def prepend_server(self, name):
        if not self.server:
            return name
        return self.server + "/" + name

    def try_pull_image(self, local_name, prepend_server):
        self.login()

        remote_name = local_name
        if prepend_server:
            remote_name = self.prepend_server(local_name)

        try:
            subprocess.run(["docker", "pull", remote_name], check=True)
        except Exception as e:
            return False

        if prepend_server:
            subprocess.run(["docker", "tag", remote_name, local_name], check=True)

        return True

    def push_image(self, local_name, prepend_server):
        self.login()

        remote_name = local_name
        if prepend_server:
            remote_name = self.prepend_server(local_name)

            subprocess.run(["docker", "tag", local_name, remote_name], check=True)

        subprocess.run(["docker", "push", remote_name], check=True)

    def image(self, *args, **kwargs):
        """
        Constructs a DockerImage that uses this index
        """
        return DockerImage(*args, registry=self, **kwargs)



import subprocess
from dataclasses import dataclass, field
from datetime import datetime
from os import PathLike
from pathlib import Path
from typing import List, Optional, Union

from .filepolicy import FilePolicy
from .hasher import Hasher
from .utils import IntervalOffset


@dataclass
class BuildConfig:
    """
    Stores all data needed to build an image. All inputs to a build should be contained here.

    A hash of all inputs can be produced - if two builds have the same hash they have the same inputs
    and thus should produce the same image.

    Params:
    dockerfile: The dockerfile to build with
    build_args: Docker build_args
    parents: List of images that this image depends on
    files: A FilePolicy object describing what files are to be made available to the build
    metadata: Additional metadata to include in the hash
    interval: An interval to refresh the hash after. For example, if you want to force a re-build every day set this interval to one day
    directory: Directory to set the build context to. Leave as None for the current directory
    unhashed_build_args: Docker build_args that won't be included in the hash. These could include credentials and other data that is required by the build
        but won't affect the built image.
    """

    dockerfile: str = "Dockerfile"
    build_args: dict = field(default_factory=dict)
    parents: List["DockerImage"] = field(default_factory=list)
    files: FilePolicy = FilePolicy.All
    metadata: str = ""
    interval: Optional[IntervalOffset] = None
    directory: Union[None, str, PathLike] = None
    unhashed_build_args: dict = field(default_factory=dict)

    def __post_init__(self):
        self.directory = Path(self.directory) if self.directory else None

    def is_hashable(self):
        if self.files == FilePolicy.All or self.files == FilePolicy.AllBut:
            return False

        return True

    def get_relative(self, path):
        if self.directory is None:
            return path

        return self.directory / path

    def add_files_to_hash(self, hasher):
        """Adds the files specified by the file policy to the state hash"""

        assert self.files != FilePolicy.All
        assert self.files != FilePolicy.AllBut

        if self.files == FilePolicy.Nothing:
            return

        if type(self.files) == FilePolicy.Only:
            for file in self.files.exceptions:
                hasher.add_file(self.get_relative(file))

    def get_hash(self):
        """Returns a hash of all build state"""

        hasher = Hasher()
        hasher.add_file(self.get_relative(self.dockerfile))
        for arg, value in self.build_args.items():
            hasher.add_str(arg)
            hasher.add_str(value)
        for parent in self.parents:
            hasher.add_str(parent.name)

        self.add_files_to_hash(hasher)

        hasher.add_str(self.metadata)
        if self.interval:
            hasher.add_str(str(self.interval.get_intervals()))

        return hasher.hexdigest()

    def create_docker_ignore_file(self):
        """
        Create a dockerignore file that either ignores everything but the given dependencies
        or ignores the given exclude paths.
        """

        lines = []
        policy_class = type(self.files)
        if self.files == FilePolicy.Nothing:
            lines = ["**"]
        elif policy_class == FilePolicy.Only:
            lines = ["**"] + [f"!{path}" for path in self.files.exceptions]
        elif policy_class == FilePolicy.AllBut:
            lines = self.files.exceptions
        elif self.files == FilePolicy.All:
            lines = []

        with open(self.get_relative(".dockerignore"), "w") as ignore_file:
            ignore_file.write("\n".join(lines))

    def build_image(self, name):
        """
        Builds an image with the config contained in this class.

        Parent images that this image depends on will be prepared first.
        """
        for parent in self.parents:
            parent.ensure()

        self.create_docker_ignore_file()

        args = ["docker", "build", "-t", name, ".", "-f", self.dockerfile]
        for arg, value in {**self.unhashed_build_args, **self.build_args}.items():
            args.extend(["--build-arg", f"{arg}={value}"])

        subprocess.run(args, check=True, cwd=self.directory)



from abc import ABC
from dataclasses import dataclass, field
from enum import Enum
from typing import List


class All:
    pass


class Nothing:
    pass


@dataclass
class AllBut:
    exceptions: List[str] = field(default_factory=list)


@dataclass
class Only:
    exceptions: List[str] = field(default_factory=list)


class FilePolicy(ABC):
    All = All
    Nothing = Nothing
    AllBut = AllBut
    Only = Only



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dockerensure']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'dockerensure',
    'version': '0.0.3',
    'description': 'Ensure that Docker images are ready for use',
    'long_description': '# Docker Ensure\n\nDocker Ensure is a tool that ensures Docker images are available on your system by building or downloading the image if necessary.\n\n## Installing Docker Ensure\n\n```\npip install dockerensure\n```\n\n## Basic Usage\n\nSay we have a Dockerfile `first.Dockerfile` that we build into `first-image`. The following code will ensure that an image called `first-image` exists by building it if it doesn\'t yet exist.\n\nThe `BuildConfig` will not make any files available to the build context by using a wildcard `.dockerignore` file.\n\n```\nDockerImage(\n    "first-image",\n    build_config=BuildConfig(\n        "first.Dockerfile",\n        files=FilePolicy.Nothing\n    )\n).ensure()\n```\n\nThe simple declaration above will only rebuild if `first-image` doesn\'t exist on the system. If we have an image registry we can include that as a place to search for images:\n\n```\nregistry = DockerRegistry("my.registry")\nDockerImage(\n    "first-image",\n    build_config=BuildConfig(\n        "first.Dockerfile",\n        files=FilePolicy.Nothing\n    ),\n    registry=registry\n).ensure()\n```\n\nDockerEnsure will check the registry for the image before building. If the image doesn\'t exist the build will take place and the image will be pushed to the registry.\nYou can control how the registry is used with the `remote_policy` argument.\n\nIn the above examples if the Dockerfile is changed the image won\'t be rebuild as long as it exists. We can include a hash in the image tag consisting of the state of the build config to detect any changes to that config:\n\n```\nimage = DockerImage(\n    "first-image",\n    with_hash=True,\n    build_config=BuildConfig(\n        "first.Dockerfile",\n        files=FilePolicy.Nothing\n    )\n)\nprint(image.ref)  # e.g. first-image:abcd9876\n```\n\nIf the contents of `first.Dockerfile` change the hash will also change and a new image will be built. In these situations the full reference of the image is not known in advance and you must use `.ref` to return the name and tag so you can use the image elsewhere.\n\n### Other build config parameters\n\nThe build config object contains all inputs to your build. You can pass a number of different parameters to this object:\n- dockerfile: The dockerfile to build with\n- build_args: Docker build args that will be passed to the build.\n- parents: List of DockerImages that this image depends on. These images will be ensured before this image is built.\n- files: A FilePolicy object describing what files are to be made available to the build\n- metadata: Additional metadata to include in the hash. This won\'t be passed to the Docker build\n- interval: An interval to refresh the hash after. \n- directory: Directory to set the build context to. Leave as None for the current directory\n- unhashed_build_args: Docker build_args that won\'t be included in the hash. These could include credentials and other data that is required by the build but won\'t affect the built image.\n\n#### File Policy\n\nYou can control what files are available to the build context by using a `FilePolicy`. `FilePolicy.All` is the default and makes all files available. `FilePolicy.Nothing` is the opposite - all files will be excluded. If you ran `COPY . .` in your Dockerfile, nothing would be copied.\n\n`FilePolicy.Only` specifies certain files to include - everything else is excluded. This is used when you know exactly which files your build needs. For example, if you\'re running `pip install -r requirements.txt` in your Dockerfile your file policy would be `FilePolicy.Only(["requirements.txt"])`.\n\n`FilePolicy.AllBut` excludes certain files. Maybe you\'re building from a Git repo - in that case use `FilePolicy.AllBut([".git/"])`.\n\nIf you are including the hash of the build state in your image name you are limited to using `Nothing` or `Only` file policies. This is because all the files included in the file policy are used to calculate the hash. Furthermore when using `Only` you must use files, not directories.\n\n#### Metadata\n\nYou can include extra string metadata that will affect the hash but not be included in the build. This is completely arbitrary and user defined.\n\n#### Interval\n\nTo prevent builds becoming outdated you can use the interval parameter to force a rebuild after a period of time. Rhis is recommended if your Dockerfile installs from a repository (e.g. `apt get update && apt get upgrade`).\n\n`IntervalOffset(timedelta(days=1))` will change the hash every day.\n\nNote that you must be using `with_hash=True` on the DockerImage for this to have any effect.\n\n#### Unhashed build args\n\nYou may want to pass build args to the build without affecting the build state hash. For example, temporary credentials that are discarded after use. In this case we need the build to have these credentials but we don\'t want the hash to change every time the credentials are rotated.\n\n`BuildConfig(..., unhashed_build_args={"REPO_TOKEN": "abcxyz"})`\n\n## Versioning Images\n\n\n\n## Images without BuildConfig\n\n',
    'author': 'ubuntom',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ubuntom/dockerensure',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
